# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DQN agent implemented in PyTorch."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import math
import numpy as np
from scipy import stats
import torch
import torch.nn as nn
import torch.nn.functional as F

from open_spiel.python import rl_agent
from open_spiel.python.utils.replay_buffer import ReplayBuffer

Transition = collections.namedtuple(
    "Transition",
    "info_state action reward next_info_state is_final_step legal_actions_mask")

ILLEGAL_ACTION_LOGITS_PENALTY = -1e9


class SonnetLinear(nn.Module):
  """A Sonnet linear module.

  Always includes biases and only supports ReLU activations.
  """

  def __init__(self, in_size, out_size, activate_relu=True):
    """Creates a Sonnet linear layer.

    Args:
      in_size: (int) number of inputs
      out_size: (int) number of outputs
      activate_relu: (bool) whether to include a ReLU activation layer
    """
    super(SonnetLinear, self).__init__()
    self._activate_relu = activate_relu
    stddev = 1.0 / math.sqrt(in_size)
    mean = 0
    lower = (-2 * stddev - mean) / stddev
    upper = (2 * stddev - mean) / stddev
    # Weight initialization inspired by Sonnet's Linear layer,
    # which cites https://arxiv.org/abs/1502.03167v3
    # pytorch default: initialized from
    # uniform(-sqrt(1/in_features), sqrt(1/in_features))
    self._weight = nn.Parameter(
        torch.Tensor(
            stats.truncnorm.rvs(
                lower, upper, loc=mean, scale=stddev, size=[out_size,
                                                            in_size])))
    self._bias = nn.Parameter(torch.zeros([out_size]))

  def forward(self, tensor):
    y = F.linear(tensor, self._weight, self._bias)
    return F.relu(y) if self._activate_relu else y


class MLP(nn.Module):
  """A simple network built from nn.linear layers."""

  def __init__(self,
               input_size,
               hidden_sizes,
               output_size,
               activate_final=False):
    """Create the MLP.

    Args:
      input_size: (int) number of inputs
      hidden_sizes: (list) sizes (number of units) of each hidden layer
      output_size: (int) number of outputs
      activate_final: (bool) should final layer should include a ReLU
    """

    super(MLP, self).__init__()
    self._layers = []
    # Hidden layers
    for size in hidden_sizes:
      self._layers.append(SonnetLinear(in_size=input_size, out_size=size))
      input_size = size
    # Output layer
    self._layers.append(
        SonnetLinear(
            in_size=input_size,
            out_size=output_size,
            activate_relu=activate_final))

    self.model = nn.ModuleList(self._layers)

  def forward(self, x):
    for layer in self.model:
      x = layer(x)
    return x


class DQN(rl_agent.AbstractAgent):
  """DQN Agent implementation in PyTorch.

  See open_spiel/python/examples/breakthrough_dqn.py for an usage example.
  """

  def __init__(self,
               player_id,
               state_representation_size,
               num_actions,
               hidden_layers_sizes=128,
               replay_buffer_capacity=10000,
               batch_size=128,
               replay_buffer_class=ReplayBuffer,
               learning_rate=0.01,
               update_target_network_every=1000,
               learn_every=10,
               discount_factor=1.0,
               min_buffer_size_to_learn=1000,
               epsilon_start=1.0,
               epsilon_end=0.1,
               epsilon_decay_duration=int(1e6),
               optimizer_str="sgd",
               loss_str="mse"):
    """Initialize the DQN agent."""

    # This call to locals() is used to store every argument used to initialize
    # the class instance, so it can be copied with no hyperparameter change.
    self._kwargs = locals()

    self.player_id = player_id
    self._num_actions = num_actions
    if isinstance(hidden_layers_sizes, int):
      hidden_layers_sizes = [hidden_layers_sizes]
    self._layer_sizes = hidden_layers_sizes
    self._batch_size = batch_size
    self._update_target_network_every = update_target_network_every
    self._learn_every = learn_every
    self._min_buffer_size_to_learn = min_buffer_size_to_learn
    self._discount_factor = discount_factor

    self._epsilon_start = epsilon_start
    self._epsilon_end = epsilon_end
    self._epsilon_decay_duration = epsilon_decay_duration

    # TODO(author6) Allow for optional replay buffer config.
    if not isinstance(replay_buffer_capacity, int):
      raise ValueError("Replay buffer capacity not an integer.")
    self._replay_buffer = replay_buffer_class(replay_buffer_capacity)
    self._prev_timestep = None
    self._prev_action = None

    # Step counter to keep track of learning, eps decay and target network.
    self._step_counter = 0

    # Keep track of the last training loss achieved in an update step.
    self._last_loss_value = None

    # Create the Q-network instances
    self._q_network = MLP(state_representation_size, self._layer_sizes,
                          num_actions)

    self._target_q_network = MLP(state_representation_size, self._layer_sizes,
                                 num_actions)

    if loss_str == "mse":
      self.loss_class = F.mse_loss
    elif loss_str == "huber":
      self.loss_class = F.smooth_l1_loss
    else:
      raise ValueError("Not implemented, choose from 'mse', 'huber'.")

    if optimizer_str == "adam":
      self._optimizer = torch.optim.Adam(
          self._q_network.parameters(), lr=learning_rate)
    elif optimizer_str == "sgd":
      self._optimizer = torch.optim.SGD(
          self._q_network.parameters(), lr=learning_rate)
    else:
      raise ValueError("Not implemented, choose from 'adam' and 'sgd'.")

  def step(self, time_step, is_evaluation=False, add_transition_record=True):
    """Returns the action to be taken and updates the Q-network if needed.

    Args:
      time_step: an instance of rl_environment.TimeStep.
      is_evaluation: bool, whether this is a training or evaluation call.
      add_transition_record: Whether to add to the replay buffer on this step.

    Returns:
      A `rl_agent.StepOutput` containing the action probs and chosen action.
    """

    # Act step: don't act at terminal info states or if its not our turn.
    if (not time_step.last()) and (
        time_step.is_simultaneous_move() or
        self.player_id == time_step.current_player()):
      info_state = time_step.observations["info_state"][self.player_id]
      legal_actions = time_step.observations["legal_actions"][self.player_id]
      epsilon = self._get_epsilon(is_evaluation)
      action, probs = self._epsilon_greedy(info_state, legal_actions, epsilon)
    else:
      action = None
      probs = []

    # Don't mess up with the state during evaluation.
    if not is_evaluation:
      self._step_counter += 1

      if self._step_counter % self._learn_every == 0:
        self._last_loss_value = self.learn()

      if self._step_counter % self._update_target_network_every == 0:
        # state_dict method returns a dictionary containing a whole state of the
        # module.
        self._target_q_network.load_state_dict(self._q_network.state_dict())

      if self._prev_timestep and add_transition_record:
        # We may omit record adding here if it's done elsewhere.
        self.add_transition(self._prev_timestep, self._prev_action, time_step)

      if time_step.last():  # prepare for the next episode.
        self._prev_timestep = None
        self._prev_action = None
        return
      else:
        self._prev_timestep = time_step
        self._prev_action = action

    return rl_agent.StepOutput(action=action, probs=probs)

  def add_transition(self, prev_time_step, prev_action, time_step):
    """Adds the new transition using `time_step` to the replay buffer.

    Adds the transition from `self._prev_timestep` to `time_step` by
    `self._prev_action`.

    Args:
      prev_time_step: prev ts, an instance of rl_environment.TimeStep.
      prev_action: int, action taken at `prev_time_step`.
      time_step: current ts, an instance of rl_environment.TimeStep.
    """
    assert prev_time_step is not None
    legal_actions = (time_step.observations["legal_actions"][self.player_id])
    legal_actions_mask = np.zeros(self._num_actions)
    legal_actions_mask[legal_actions] = 1.0
    transition = Transition(
        info_state=(
            prev_time_step.observations["info_state"][self.player_id][:]),
        action=prev_action,
        reward=time_step.rewards[self.player_id],
        next_info_state=time_step.observations["info_state"][self.player_id][:],
        is_final_step=float(time_step.last()),
        legal_actions_mask=legal_actions_mask)
    self._replay_buffer.add(transition)

  def _epsilon_greedy(self, info_state, legal_actions, epsilon):
    """Returns a valid epsilon-greedy action and valid action probs.

    Action probabilities are given by a softmax over legal q-values.

    Args:
      info_state: hashable representation of the information state.
      legal_actions: list of legal actions at `info_state`.
      epsilon: float, probability of taking an exploratory action.

    Returns:
      A valid epsilon-greedy action and valid action probabilities.
    """
    probs = np.zeros(self._num_actions)
    if np.random.rand() < epsilon:
      action = np.random.choice(legal_actions)
      probs[legal_actions] = 1.0 / len(legal_actions)
    else:
      info_state = torch.Tensor(np.reshape(info_state, [1, -1]))
      q_values = self._q_network(info_state).detach()[0]
      legal_q_values = q_values[legal_actions]
      action = legal_actions[torch.argmax(legal_q_values)]
      probs[action] = 1.0
    return action, probs

  def _get_epsilon(self, is_evaluation, power=1.0):
    """Returns the evaluation or decayed epsilon value."""
    if is_evaluation:
      return 0.0
    decay_steps = min(self._step_counter, self._epsilon_decay_duration)
    decayed_epsilon = (
        self._epsilon_end + (self._epsilon_start - self._epsilon_end) *
        (1 - decay_steps / self._epsilon_decay_duration)**power)
    return decayed_epsilon

  def learn(self):
    """Compute the loss on sampled transitions and perform a Q-network update.

    If there are not enough elements in the buffer, no loss is computed and
    `None` is returned instead.

    Returns:
      The average loss obtained on this batch of transitions or `None`.
    """

    if (len(self._replay_buffer) < self._batch_size or
        len(self._replay_buffer) < self._min_buffer_size_to_learn):
      return None

    transitions = self._replay_buffer.sample(self._batch_size)
    info_states = torch.Tensor([t.info_state for t in transitions])
    actions = torch.LongTensor([t.action for t in transitions])
    rewards = torch.Tensor([t.reward for t in transitions])
    next_info_states = torch.Tensor([t.next_info_state for t in transitions])
    are_final_steps = torch.Tensor([t.is_final_step for t in transitions])
    legal_actions_mask = torch.Tensor(
        [t.legal_actions_mask for t in transitions])

    self._q_values = self._q_network(info_states)
    self._target_q_values = self._target_q_network(next_info_states).detach()

    illegal_actions = 1 - legal_actions_mask
    illegal_logits = illegal_actions * ILLEGAL_ACTION_LOGITS_PENALTY
    max_next_q = torch.max(self._target_q_values + illegal_logits, dim=1)[0]
    target = (
        rewards + (1 - are_final_steps) * self._discount_factor * max_next_q)
    action_indices = torch.stack([
        torch.arange(self._q_values.shape[0], dtype=torch.long), actions
    ],
                                 dim=0)
    predictions = self._q_values[list(action_indices)]

    loss = self.loss_class(predictions, target)

    self._optimizer.zero_grad()
    loss.backward()
    self._optimizer.step()

    return loss

  @property
  def q_values(self):
    return self._q_values

  @property
  def replay_buffer(self):
    return self._replay_buffer

  @property
  def loss(self):
    return self._last_loss_value

  @property
  def prev_timestep(self):
    return self._prev_timestep

  @property
  def prev_action(self):
    return self._prev_action

  @property
  def step_counter(self):
    return self._step_counter

  def get_weights(self):
    variables = [m.weight for m in self._q_network.model]
    variables.append([m.weight for m in self._target_q_network.model])
    return variables

  def copy_with_noise(self, sigma=0.0, copy_weights=True):
    """Copies the object and perturbates it with noise.

    Args:
      sigma: gaussian dropout variance term : Multiplicative noise following
        (1+sigma*epsilon), epsilon standard gaussian variable, multiplies each
        model weight. sigma=0 means no perturbation.
      copy_weights: Boolean determining whether to copy model weights (True) or
        just model hyperparameters.

    Returns:
      Perturbated copy of the model.
    """
    _ = self._kwargs.pop("self", None)
    copied_object = DQN(**self._kwargs)

    q_network = getattr(copied_object, "_q_network")
    target_q_network = getattr(copied_object, "_target_q_network")

    if copy_weights:
      with torch.no_grad():
        for q_model in q_network.model:
          q_model.weight *= (1 + sigma * torch.randn(q_model.weight.shape))
        for tq_model in target_q_network.model:
          tq_model.weight *= (1 + sigma * torch.randn(tq_model.weight.shape))
    return copied_object



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Boltzmann DQN agent implemented in JAX.

This algorithm is a variation of DQN that uses a softmax policy directly with
the unregularized action-value function. See https://arxiv.org/abs/2102.01585.
"""

import jax
import jax.numpy as jnp
import numpy as np

from open_spiel.python.jax import dqn


class BoltzmannDQN(dqn.DQN):
  """Boltzmann DQN implementation in JAX."""

  def __init__(self, *args, eta: float = 1.0, seed: int = 42, **kwargs):
    """Initializes the Boltzmann DQN agent.

    Args:
      *args: args passed to the underlying DQN agent.
      eta: Temperature parameter used in the softmax function.
      seed: Random seed used for action selection.
      **kwargs: kwargs passed to the underlying DQN agent.
    """
    self._eta = eta
    self._rs = np.random.RandomState(seed)  # Used to select actions.
    super().__init__(*args, seed=seed, **kwargs)

  def _create_networks(self, rng, state_representation_size):
    """Called to create the networks."""
    # We use the DQN networks and an additional network for the fixed policy.
    super()._create_networks(rng, state_representation_size)
    self.params_prev_q_network = self.hk_network.init(
        rng, jnp.ones([1, state_representation_size]))

  def _softmax_action_probs(self,
                            params,
                            info_state,
                            legal_actions,
                            coeff=None):
    """Returns a valid soft-max action and action probabilities.

    Args:
      params: Parameters of the Q-network.
      info_state: Observations from the environment.
      legal_actions: List of legal actions.
      coeff: If not None, then the terms in softmax function will be
        element-wise multiplied with these coefficients.

    Returns:
      a valid soft-max action and action probabilities.
    """
    info_state = np.reshape(info_state, [1, -1])
    q_values = self.hk_network_apply(params, info_state)[0]
    legal_one_hot = self._to_one_hot(legal_actions)
    legal_q_values = (
        q_values + (1 - legal_one_hot) * dqn.ILLEGAL_ACTION_LOGITS_PENALTY)
    # Apply temperature and subtract the maximum value for numerical stability.
    temp = legal_q_values / self._eta
    unnormalized = np.exp(temp - np.amax(temp))
    if coeff is not None:
      unnormalized = np.multiply(coeff, unnormalized)
    probs = unnormalized / unnormalized.sum()
    action = self._rs.choice(legal_actions, p=probs[legal_actions])
    return action, probs

  def _get_action_probs(self, info_state, legal_actions, is_evaluation=False):
    """Returns a selected action and the probabilities of legal actions."""
    if is_evaluation:
      # Soft-max normalized by the action probabilities from the previous
      # Q-network.
      _, prev_probs = self._softmax_action_probs(self.params_prev_q_network,
                                                 info_state, legal_actions)
      return self._softmax_action_probs(self.params_q_network, info_state,
                                        legal_actions, prev_probs)

    # During training, we use the DQN action selection, which will be
    # epsilon-greedy.
    return super()._get_action_probs(
        info_state, legal_actions, is_evaluation=False)

  def update_prev_q_network(self):
    """Updates the parameters of the previous Q-network."""
    self.params_prev_q_network = jax.tree_multimap(lambda x: x.copy(),
                                                   self.params_q_network)



#!/usr/bin/env python3
#
# Copyright 2021 The Abseil Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A script to do source transformations to create a new LTS release.

   Usage: ./create_lts.py YYYYMMDD
"""

import sys


def ReplaceStringsInFile(filename, replacement_dict):
  """Performs textual replacements in a file.

  Rewrites filename with the keys in replacement_dict replaced with
  their values. This function assumes the file can fit in memory.

  Args:
    filename: the filename to perform the replacement on
    replacement_dict: a dictionary of key strings to be replaced with their
      values

  Raises:
    Exception: A failure occured
  """
  f = open(filename, 'r')
  content = f.read()
  f.close()

  for key, value in replacement_dict.items():
    original = content
    content = content.replace(key, value)
    if content == original:
      raise Exception('Failed to find {} in {}'.format(key, filename))

  f = open(filename, 'w')
  f.write(content)
  f.close()


def StripContentBetweenTags(filename, strip_begin_tag, strip_end_tag):
  """Strip contents from a file.

  Rewrites filename with by removing all content between
  strip_begin_tag and strip_end_tag, including the tags themselves.

  Args:
    filename: the filename to perform the replacement on
    strip_begin_tag: the start of the content to be removed
    strip_end_tag: the end of the content to be removed

  Raises:
    Exception: A failure occured
  """
  f = open(filename, 'r')
  content = f.read()
  f.close()

  while True:
    begin = content.find(strip_begin_tag)
    if begin == -1:
      break
    end = content.find(strip_end_tag, begin + len(strip_begin_tag))
    if end == -1:
      raise Exception('{}: imbalanced strip begin ({}) and '
                      'end ({}) tags'.format(filename, strip_begin_tag,
                                             strip_end_tag))
    content = content.replace(content[begin:end + len(strip_end_tag)], '')

  f = open(filename, 'w')
  f.write(content)
  f.close()


def main(argv):
  if len(argv) != 2:
    print('Usage: {} YYYYMMDD'.format(sys.argv[0], file=sys.stderr))
    sys.exit(1)

  datestamp = sys.argv[1]
  if len(datestamp) != 8 or not datestamp.isdigit():
    raise Exception(
        'datestamp={} is not in the YYYYMMDD format'.format(datestamp))

  # Replacement directives go here.
  ReplaceStringsInFile(
      'absl/base/config.h', {
          '#undef ABSL_LTS_RELEASE_VERSION':
              '#define ABSL_LTS_RELEASE_VERSION {}'.format(datestamp),
          '#undef ABSL_LTS_RELEASE_PATCH_LEVEL':
              '#define ABSL_LTS_RELEASE_PATCH_LEVEL 0'
      })
  ReplaceStringsInFile(
      'absl/base/options.h', {
          '#define ABSL_OPTION_USE_INLINE_NAMESPACE 0':
              '#define ABSL_OPTION_USE_INLINE_NAMESPACE 1',
          '#define ABSL_OPTION_INLINE_NAMESPACE_NAME head':
              '#define ABSL_OPTION_INLINE_NAMESPACE_NAME lts_{}'.format(
                  datestamp)
      })
  ReplaceStringsInFile(
      'CMakeLists.txt', {
          'project(absl LANGUAGES CXX)':
              'project(absl LANGUAGES CXX VERSION {})'.format(datestamp)
      })
  # Set the SOVERSION to YYMM.0.0 - The first 0 means we only have ABI
  # compatible changes, and the second 0 means we can increment it to
  # mark changes as ABI-compatible, for patch releases.  Note that we
  # only use the last two digits of the year and the month because the
  # MacOS linker requires the first part of the SOVERSION to fit into
  # 16 bits.
  # https://www.sicpers.info/2013/03/how-to-version-a-mach-o-library/
  ReplaceStringsInFile(
      'CMake/AbseilHelpers.cmake',
      {'SOVERSION 0': 'SOVERSION "{}.0.0"'.format(datestamp[2:6])})
  StripContentBetweenTags('CMakeLists.txt', '# absl:lts-remove-begin',
                          '# absl:lts-remove-end')


if __name__ == '__main__':
  main(sys.argv)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.algorithms.policy_value."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import absltest
import numpy as np

from open_spiel.python import games  # pylint: disable=unused-import
from open_spiel.python import policy
from open_spiel.python.algorithms import expected_game_score
import pyspiel


class PolicyValueTest(absltest.TestCase):

  def test_expected_game_score_uniform_random_kuhn_poker(self):
    game = pyspiel.load_game("kuhn_poker")
    uniform_policy = policy.UniformRandomPolicy(game)
    uniform_policy_values = expected_game_score.policy_value(
        game.new_initial_state(), [uniform_policy] * 2)
    self.assertTrue(np.allclose(uniform_policy_values, [1 / 8, -1 / 8]))

  def test_expected_game_score_uniform_random_iterated_prisoner_dilemma(self):
    game = pyspiel.load_game(
        "python_iterated_prisoners_dilemma(max_game_length=6)")
    pi = policy.UniformRandomPolicy(game)
    values = expected_game_score.policy_value(game.new_initial_state(), pi)
    # 4*(1-0.875**6)/0.125 = 17.6385498
    np.testing.assert_allclose(values, [17.6385498, 17.6385498])


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.algorithms.exploitability."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import absltest
from absl.testing import parameterized

from open_spiel.python import policy
from open_spiel.python.algorithms import exploitability
from open_spiel.python.algorithms import policy_utils
from open_spiel.python.games import data
import pyspiel


class ExploitabilityTest(parameterized.TestCase):

  def test_exploitability_on_kuhn_poker_uniform_random(self):
    # NashConv of uniform random test_policy from (found on Google books):
    # https://link.springer.com/chapter/10.1007/978-3-319-75931-9_5
    game = pyspiel.load_game("kuhn_poker")
    test_policy = policy.UniformRandomPolicy(game)
    expected_nash_conv = 11 / 12
    self.assertAlmostEqual(
        exploitability.exploitability(game, test_policy),
        expected_nash_conv / 2)

  def test_kuhn_poker_uniform_random_best_response_pid0(self):
    game = pyspiel.load_game("kuhn_poker")
    test_policy = policy.UniformRandomPolicy(game)
    results = exploitability.best_response(game, test_policy, player_id=0)
    self.assertEqual(
        results["best_response_action"],
        {
            "0": 1,  # Bet in case opponent folds when winning
            "1": 1,  # Bet in case opponent folds when winning
            "2": 0,  # Both equally good (we return the lowest action)
            # Some of these will never happen under the best-response policy,
            # but we have computed best-response actions anyway.
            "0pb": 0,  # Fold - we're losing
            "1pb": 1,  # Call - we're 50-50
            "2pb": 1,  # Call - we've won
        })
    self.assertGreater(results["nash_conv"], 0.1)

  def test_kuhn_poker_uniform_random_best_response_pid1(self):
    game = pyspiel.load_game("kuhn_poker")
    test_policy = policy.UniformRandomPolicy(game)
    results = exploitability.best_response(game, test_policy, player_id=1)
    self.assertEqual(
        results["best_response_action"],
        {
            # Bet is always best
            "0p": 1,
            "1p": 1,
            "2p": 1,
            # Call unless we know we're beaten
            "0b": 0,
            "1b": 1,
            "2b": 1,
        })
    self.assertGreater(results["nash_conv"], 0.1)

  def test_kuhn_poker_uniform_random(self):
    # NashConv of uniform random test_policy from (found on Google books):
    # https://link.springer.com/chapter/10.1007/978-3-319-75931-9_5
    game = pyspiel.load_game("kuhn_poker")
    test_policy = policy.UniformRandomPolicy(game)
    self.assertAlmostEqual(exploitability.nash_conv(game, test_policy), 11 / 12)

  def test_kuhn_poker_always_fold(self):
    game = pyspiel.load_game("kuhn_poker")
    test_policy = policy.FirstActionPolicy(game)
    self.assertAlmostEqual(exploitability.nash_conv(game, test_policy), 2)

  def test_kuhn_poker_optimal(self):
    game = pyspiel.load_game("kuhn_poker")
    test_policy = data.kuhn_nash_equilibrium(alpha=0.2)
    self.assertAlmostEqual(exploitability.nash_conv(game, test_policy), 0)

  def test_leduc_poker_uniform_random(self):
    # NashConv taken from independent implementations
    game = pyspiel.load_game("leduc_poker")
    test_policy = policy.UniformRandomPolicy(game)
    self.assertAlmostEqual(
        exploitability.nash_conv(game, test_policy), 4.747222222222222)

  def test_leduc_poker_always_fold(self):
    game = pyspiel.load_game("leduc_poker")
    test_policy = policy.FirstActionPolicy(game)
    self.assertAlmostEqual(exploitability.nash_conv(game, test_policy), 2)

  # Values for uniform policies taken from
  # https://link.springer.com/chapter/10.1007/978-3-319-75931-9_5
  # (including multiplayer games below). However, the value for Leduc against
  # the uniform test_policy is wrong in the paper. This has been independently
  # verified in a number of independent code bases. The 4.7472 value is correct.
  # Value for AlwaysFold is trivial: if you
  # always fold, you win 0 chips, but if you switch to AlwaysBet, you win 1
  # chip everytime if playing against a player who always folds.
  @parameterized.parameters(
      ("kuhn_poker", policy.UniformRandomPolicy, 0.9166666666666666),
      ("kuhn_poker", policy.FirstActionPolicy, 2.),
      ("kuhn_poker", lambda _: data.kuhn_nash_equilibrium(alpha=0.2), 0.),
      ("leduc_poker", policy.FirstActionPolicy, 2.),
      ("leduc_poker", policy.UniformRandomPolicy, 4.7472222222222),
  )
  def test_2p_nash_conv(self, game_name, policy_func, expected):
    game = pyspiel.load_game(game_name)
    self.assertAlmostEqual(
        exploitability.nash_conv(game, policy_func(game)), expected)

  @parameterized.parameters(3, 4)
  def test_kuhn_poker_uniform_random_nash_conv(self, num_players):
    game = pyspiel.load_game("kuhn_poker", {"players": num_players})
    test_policy = policy.UniformRandomPolicy(game)
    self.assertGreater(exploitability.nash_conv(game, test_policy), 2)

  @parameterized.parameters(("kuhn_poker", 2), ("kuhn_poker", 3),
                            ("kuhn_poker", 4))
  def test_python_same_as_cpp_for_multiplayer_uniform_random_nash_conv(
      self, game_name, num_players):
    game = pyspiel.load_game(game_name, {"players": num_players})

    # TabularPolicy defaults to being a uniform random policy.
    test_policy = policy.TabularPolicy(game)
    python_nash_conv = exploitability.nash_conv(game, test_policy)
    cpp_nash_conv = pyspiel.nash_conv(
        game, policy_utils.policy_to_dict(test_policy, game))
    self.assertAlmostEqual(python_nash_conv, cpp_nash_conv)

  def test_cpp_python_cfr_kuhn(self):
    game = pyspiel.load_game("kuhn_poker")
    solver = pyspiel.CFRSolver(game)
    for _ in range(100):
      solver.evaluate_and_update_policy()
    pyspiel_average_policy = solver.tabular_average_policy()
    cpp_nash_conv = pyspiel.nash_conv(game, pyspiel_average_policy)
    python_policy = policy.pyspiel_policy_to_python_policy(
        game, pyspiel_average_policy)
    python_nash_conv = exploitability.nash_conv(game, python_policy)
    self.assertAlmostEqual(python_nash_conv, cpp_nash_conv)

if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for open_spiel.python.utils.file_logger."""

import os
import tempfile

from absl.testing import absltest

from open_spiel.python.utils import file_logger


class FileLoggerTest(absltest.TestCase):

  def test_file_logger(self):
    tmp_dir = tempfile.mkdtemp()
    try:
      log_name = "test"
      log_file_name = os.path.join(tmp_dir, "log-{}.txt".format(log_name))

      self.assertTrue(os.path.isdir(tmp_dir))
      self.assertFalse(os.path.exists(log_file_name))

      with file_logger.FileLogger(tmp_dir, log_name) as logger:
        logger.print("line 1")
        logger.print("line", 2)
        logger.print("line", 3, "asdf")

      with open(log_file_name, "r") as f:
        lines = f.readlines()

        self.assertLen(lines, 3)
        self.assertIn("line 1", lines[0])
        self.assertIn("line 2", lines[1])
        self.assertIn("line 3 asdf", lines[2])
    finally:
      if os.path.exists(log_file_name):
        os.remove(log_file_name)
      os.rmdir(tmp_dir)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Train a policy net on bridge bidding based on a dataset of trajectories.

Suitable data for training, generated by WBridge5, may be downloaded from:
https://console.cloud.google.com/storage/browser/openspiel-data/bridge
"""

import os
import pickle
from typing import Any, Tuple

from absl import app
from absl import flags

import haiku as hk
import jax
from jax import numpy as jnp
import numpy as np
import optax

import pyspiel

OptState = Any
Params = Any

FLAGS = flags.FLAGS
GAME = pyspiel.load_game('bridge(use_double_dummy_result=false)')
NUM_ACTIONS = 38
MIN_ACTION = 52
NUM_CARDS = 52
NUM_PLAYERS = 4
TOP_K_ACTIONS = 5  # How many alternative actions to display

flags.DEFINE_integer('iterations', 100000, 'Number of iterations')
flags.DEFINE_string('data_path', None, 'Location for data')
flags.DEFINE_integer('eval_every', 10000, 'How often to evaluate the policy')
flags.DEFINE_integer('num_examples', 3,
                     'How many examples to print per evaluation')
flags.DEFINE_integer('train_batch', 128, 'Batch size for training step')
flags.DEFINE_integer('eval_batch', 10000, 'Batch size when evaluating')
flags.DEFINE_integer('rng_seed', 42, 'Seed for initial network weights')
flags.DEFINE_string('save_path', None, 'Location for saved networks')


def _no_play_trajectory(line: str):
  """Returns the deal and bidding actions only given a text trajectory."""
  actions = [int(x) for x in line.split(' ')]
  # Usually a trajectory is NUM_CARDS chance events for the deal, plus one
  # action for every bid of the auction, plus NUM_CARDS actions for the play
  # phase. Exceptionally, if all NUM_PLAYERS players Pass, there is no play
  # phase and the trajectory is just of length NUM_CARDS + NUM_PLAYERS.
  if len(actions) == NUM_CARDS + NUM_PLAYERS:
    return tuple(actions)
  else:
    return tuple(actions[:-NUM_CARDS])


def make_dataset(file: str):
  """Creates dataset as a generator of single examples."""
  all_trajectories = [_no_play_trajectory(line) for line in open(file)]
  while True:
    np.random.shuffle(all_trajectories)
    for trajectory in all_trajectories:
      action_index = np.random.randint(52, len(trajectory))
      state = GAME.new_initial_state()
      for action in trajectory[:action_index]:
        state.apply_action(action)
      yield (state.observation_tensor(), trajectory[action_index] - MIN_ACTION)


def batch(dataset, batch_size: int):
  """Creates a batched dataset from a one-at-a-time dataset."""
  observations = np.zeros([batch_size] + GAME.observation_tensor_shape(),
                          np.float32)
  labels = np.zeros(batch_size, dtype=np.int32)
  while True:
    for batch_index in range(batch_size):
      observations[batch_index], labels[batch_index] = next(dataset)
    yield observations, labels


def one_hot(x, k):
  """Returns a one-hot encoding of `x` of size `k`."""
  return jnp.array(x[..., jnp.newaxis] == jnp.arange(k), dtype=np.float32)


def net_fn(x):
  """Haiku module for our network."""
  net = hk.Sequential([
      hk.Linear(1024),
      jax.nn.relu,
      hk.Linear(1024),
      jax.nn.relu,
      hk.Linear(1024),
      jax.nn.relu,
      hk.Linear(1024),
      jax.nn.relu,
      hk.Linear(NUM_ACTIONS),
      jax.nn.log_softmax,
  ])
  return net(x)


def main(argv):
  if len(argv) > 1:
    raise app.UsageError('Too many command-line arguments.')

  # Make the network.
  net = hk.without_apply_rng(hk.transform(net_fn))

  # Make the optimiser.
  opt = optax.adam(1e-4)

  @jax.jit
  def loss(
      params: Params,
      inputs: np.ndarray,
      targets: np.ndarray,
  ) -> jnp.DeviceArray:
    """Cross-entropy loss."""
    assert targets.dtype == np.int32
    log_probs = net.apply(params, inputs)
    return -jnp.mean(one_hot(targets, NUM_ACTIONS) * log_probs)

  @jax.jit
  def accuracy(
      params: Params,
      inputs: np.ndarray,
      targets: np.ndarray,
  ) -> jnp.DeviceArray:
    """Classification accuracy."""
    predictions = net.apply(params, inputs)
    return jnp.mean(jnp.argmax(predictions, axis=-1) == targets)

  @jax.jit
  def update(
      params: Params,
      opt_state: OptState,
      inputs: np.ndarray,
      targets: np.ndarray,
  ) -> Tuple[Params, OptState]:
    """Learning rule (stochastic gradient descent)."""
    _, gradient = jax.value_and_grad(loss)(params, inputs, targets)
    updates, opt_state = opt.update(gradient, opt_state)
    new_params = optax.apply_updates(params, updates)
    return new_params, opt_state

  def output_samples(params: Params, max_samples: int):
    """Output some cases where the policy disagrees with the dataset action."""
    if max_samples == 0:
      return
    count = 0
    with open(os.path.join(FLAGS.data_path, 'test.txt')) as f:
      lines = list(f)
    np.random.shuffle(lines)
    for line in lines:
      state = GAME.new_initial_state()
      actions = _no_play_trajectory(line)
      for action in actions:
        if not state.is_chance_node():
          observation = np.array(state.observation_tensor(), np.float32)
          policy = np.exp(net.apply(params, observation))
          probs_actions = [(p, a + MIN_ACTION) for a, p in enumerate(policy)]
          pred = max(probs_actions)[1]
          if pred != action:
            print(state)
            for p, a in reversed(sorted(probs_actions)[-TOP_K_ACTIONS:]):
              print('{:7} {:.2f}'.format(state.action_to_string(a), p))
            print('Ground truth {}\n'.format(state.action_to_string(action)))
            count += 1
            break
        state.apply_action(action)
      if count >= max_samples:
        return

  # Make datasets.
  if FLAGS.data_path is None:
    raise app.UsageError(
        'Please generate your own supervised training data or download from '
        'https://console.cloud.google.com/storage/browser/openspiel-data/bridge'
        ' and supply the local location as --data_path')
  train = batch(
      make_dataset(os.path.join(FLAGS.data_path, 'train.txt')),
      FLAGS.train_batch)
  test = batch(
      make_dataset(os.path.join(FLAGS.data_path, 'test.txt')), FLAGS.eval_batch)

  # Initialize network and optimiser.
  rng = jax.random.PRNGKey(FLAGS.rng_seed)  # seed used for network weights
  inputs, unused_targets = next(train)
  params = net.init(rng, inputs)
  opt_state = opt.init(params)

  # Train/eval loop.
  for step in range(FLAGS.iterations):
    # Do SGD on a batch of training examples.
    inputs, targets = next(train)
    params, opt_state = update(params, opt_state, inputs, targets)

    # Periodically evaluate classification accuracy on the test set.
    if (1 + step) % FLAGS.eval_every == 0:
      inputs, targets = next(test)
      test_accuracy = accuracy(params, inputs, targets)
      print(f'After {1+step} steps, test accuracy: {test_accuracy}.')
      if FLAGS.save_path:
        filename = os.path.join(FLAGS.save_path, f'params-{1 + step}.pkl')
        with open(filename, 'wb') as pkl_file:
          pickle.dump(params, pkl_file)
      output_samples(params, FLAGS.num_examples)


if __name__ == '__main__':
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Catch reinforcement learning environment."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import numpy as np

from open_spiel.python import rl_environment

# Actions
NOOP = 0
LEFT = 1
RIGHT = 2

_Point = collections.namedtuple("Point", ["x", "y"])


class Environment(object):
  """A catch reinforcement learning environment.

  The implementation considers illegal actions: trying to move the paddle in the
  wall direction when next to a wall will incur in an invalid action and an
  error will be purposely raised.
  """

  def __init__(self, discount=1.0, width=5, height=10, seed=None):
    self._rng = np.random.RandomState(seed)
    self._width = width
    self._height = height
    self._should_reset = True
    self._num_actions = 3

    # Discount returned at non-initial steps.
    self._discounts = [discount] * self.num_players

  def reset(self):
    """Resets the environment."""
    self._should_reset = False
    self._ball_pos = _Point(x=self._rng.randint(0, self._width - 1), y=0)
    self._paddle_pos = _Point(
        x=self._rng.randint(0, self._width - 1), y=self._height - 1)

    legal_actions = [NOOP]
    if self._paddle_pos.x > 0:
      legal_actions.append(LEFT)
    if self._paddle_pos.x < self._width - 1:
      legal_actions.append(RIGHT)

    observations = {
        "info_state": [self._get_observation()],
        "legal_actions": [legal_actions],
        "current_player": 0,
    }

    return rl_environment.TimeStep(
        observations=observations,
        rewards=None,
        discounts=None,
        step_type=rl_environment.StepType.FIRST)

  def step(self, actions):
    """Updates the environment according to `actions` and returns a `TimeStep`.

    Args:
      actions: A singleton list with an integer, or an integer, representing the
        action the agent took.

    Returns:
      A `rl_environment.TimeStep` namedtuple containing:
        observation: singleton list of dicts containing player observations,
            each corresponding to `observation_spec()`.
        reward: singleton list containing the reward at this timestep, or None
            if step_type is `rl_environment.StepType.FIRST`.
        discount: singleton list containing the discount in the range [0, 1], or
            None if step_type is `rl_environment.StepType.FIRST`.
        step_type: A `rl_environment.StepType` value.
    """
    if self._should_reset:
      return self.reset()

    if isinstance(actions, list):
      action = actions[0]
    elif isinstance(actions, int):
      action = actions
    else:
      raise ValueError("Action not supported.", actions)

    # Update paddle position
    x, y = self._paddle_pos.x, self._paddle_pos.y
    if action == LEFT:
      x -= 1
    elif action == RIGHT:
      x += 1
    elif action != NOOP:
      raise ValueError("unrecognized action ", action)

    assert 0 <= x < self._width, (
        "Illegal action detected ({}), new state: ({},{})".format(action, x, y))
    self._paddle_pos = _Point(x, y)

    # Update ball position
    x, y = self._ball_pos.x, self._ball_pos.y
    if y == self._height - 1:
      done = True
      reward = 1.0 if x == self._paddle_pos.x else -1.0
    else:
      done = False
      y += 1
      reward = 0.0
      self._ball_pos = _Point(x, y)

    # Return observation
    step_type = (
        rl_environment.StepType.LAST if done else rl_environment.StepType.MID)
    self._should_reset = step_type == rl_environment.StepType.LAST

    legal_actions = [NOOP]
    if self._paddle_pos.x > 0:
      legal_actions.append(LEFT)
    if self._paddle_pos.x < self._width - 1:
      legal_actions.append(RIGHT)

    observations = {
        "info_state": [self._get_observation()],
        "legal_actions": [legal_actions],
        "current_player": 0,
    }

    return rl_environment.TimeStep(
        observations=observations,
        rewards=[reward],
        discounts=self._discounts,
        step_type=step_type)

  def _get_observation(self):
    board = np.zeros((self._height, self._width), dtype=np.float32)
    board[self._ball_pos.y, self._ball_pos.x] = 1.0
    board[self._paddle_pos.y, self._paddle_pos.x] = 1.0
    return board.flatten()

  def observation_spec(self):
    """Defines the observation provided by the environment.

    Each dict member will contain its expected structure and shape.

    Returns:
      A specification dict describing the observation fields and shapes.
    """
    return dict(
        info_state=tuple([self._height * self._width]),
        legal_actions=(self._num_actions,),
        current_player=(),
    )

  def action_spec(self):
    """Defines action specifications.

    Specifications include action boundaries and their data type.

    Returns:
      A specification dict containing action properties.
    """
    return dict(num_actions=self._num_actions, min=0, max=2, dtype=int)

  @property
  def num_players(self):
    return 1

  @property
  def is_turn_based(self):
    return False



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of exploitability descent.

See "Computing Approximate Equilibria in Sequential Adversarial Games by
Exploitability Descent" https://arxiv.org/abs/1903.05614

The exploitability descent algorithm solves a game by repeatedly performing
the following update:

1. Construct a (deterministic) best response to our current strategy
2. Compute the value of every action in every state when playing our current
   strategy vs the best response.
3. Update our current strategy to do better vs the current best response
   by performing a policy-gradient update.

This module provides a function that returns a loss for network training, and
a Solver class that uses this loss in a tabular Exploitability Descent.

The code can be used either for a tabular exploitability descent algorithm,
as demonstrated by exploitability_descent_test, or for a neural network policy,
as in ../examples/exploitability_descent.py.

Additionally, for a minibatch version of the algorithm (which samples
uniformly across all states in the game to generate a minibatch), see the
minibatch_loss method.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import tensorflow.compat.v1 as tf

from open_spiel.python import policy
from open_spiel.python.algorithms import action_value_vs_best_response
from open_spiel.python.algorithms import masked_softmax

# Temporary disabling of v2 behavior until code is updated.
tf.disable_v2_behavior()

_NUM_PLAYERS = 2


def _create_policy_evaluator(tabular_policy, q_value_calculator):
  """Creates a function to evaluate our policy."""

  def evaluate_policy(policy_values):
    """Evaluates a tabular policy; intended to be used as a tf.py_function."""
    tabular_policy.action_probability_array = policy_values
    evaluations = [
        q_value_calculator(player, tabular_policy,
                           tabular_policy.states_per_player[player])
        for player in range(_NUM_PLAYERS)
    ]
    nash_conv = np.float64(sum([evaluations[p].exploitability for p in [0, 1]]))

    q_values = np.concatenate(
        [np.array(evaluations[p].values_vs_br, np.float64) for p in [0, 1]])
    cf_reach_probabilities = np.concatenate([
        np.array(evaluations[p].counterfactual_reach_probs_vs_br, np.float64)
        for p in [0, 1]
    ])
    return nash_conv, q_values, cf_reach_probabilities

  return evaluate_policy


class LossCalculator(object):
  """Computes the exploitability descent loss for a two-player game."""

  def __init__(self, game):
    """Initializes a loss calculation for the given game."""
    if game.num_players() != _NUM_PLAYERS:
      raise ValueError("Game {} does not have {} players.".format(
          game, _NUM_PLAYERS))
    self.tabular_policy = policy.TabularPolicy(game)
    self.q_value_calculator = action_value_vs_best_response.Calculator(game)

  def masked_softmax(self, logits):
    """Safe masked softmax."""
    return masked_softmax.tf_masked_softmax(
        logits, self.tabular_policy.legal_actions_mask)

  def loss(self, policy_values):
    """Returns the exploitability descent loss given a policy."""

    evaluate_policy = _create_policy_evaluator(self.tabular_policy,
                                               self.q_value_calculator)
    nash_conv, q_values, cf_reach_probabilities = tf.py_func(
        evaluate_policy, [policy_values], [tf.float64, tf.float64, tf.float64])
    baseline = tf.reduce_sum(policy_values * q_values, axis=-1, keepdims=True)
    advantage = q_values - tf.stop_gradient(baseline)
    loss_per_state = -tf.reduce_sum(policy_values * advantage, axis=-1)
    return nash_conv, tf.reduce_sum(loss_per_state * cf_reach_probabilities)

  def minibatch_loss(self, policy_values, q_values, indices):
    """Returns the exploitability descent loss given a policy for a subset."""

    evaluate_policy = _create_policy_evaluator(self.tabular_policy,
                                               self.q_value_calculator)
    nash_conv, real_q_values, cf_reach_probabilities = tf.py_func(
        evaluate_policy, [policy_values], [tf.float64, tf.float64, tf.float64])
    baseline = tf.reduce_sum(policy_values * q_values, axis=-1, keepdims=True)
    advantage = q_values - baseline

    # We now select a minibatch from the data to propagate our loss on.
    policy_values = tf.gather(policy_values, indices)
    advantage = tf.gather(advantage, indices)
    cf_reach_probabilities = tf.gather(cf_reach_probabilities, indices)

    # The rest is the same as before.
    loss_per_state = -tf.reduce_sum(
        policy_values * tf.stop_gradient(advantage), axis=-1)
    q_value_loss = tf.reduce_mean((q_values - real_q_values)**2, axis=1)
    q_value_loss = tf.gather(q_value_loss, indices)
    q_value_loss = tf.reduce_sum(q_value_loss * cf_reach_probabilities)
    policy_loss = tf.reduce_sum(loss_per_state * cf_reach_probabilities)
    return nash_conv, q_value_loss, policy_loss


class Solver(object):
  """Solves a two-player game using exploitability descent."""

  def __init__(self, game):
    """Initializes a solver for the given game."""
    self._loss_calculator = LossCalculator(game)
    self._logits = tf.Variable(
        np.ones_like(
            self._loss_calculator.tabular_policy.action_probability_array,
            dtype=np.float64),
        name="logits",
        use_resource=True)
    self._tabular_policy = self._loss_calculator.masked_softmax(self._logits)
    self._nash_conv, self._loss = self._loss_calculator.loss(
        self._tabular_policy)
    self._learning_rate = tf.placeholder(tf.float64, (), name="learning_rate")
    self._optimizer = tf.train.GradientDescentOptimizer(self._learning_rate)
    self._optimizer_step = self._optimizer.minimize(self._loss)

  def step(self, session, learning_rate):
    """Takes a single exploitability descent step."""
    _, nash_conv = session.run([self._optimizer_step, self._nash_conv],
                               feed_dict={self._learning_rate: learning_rate})
    return nash_conv



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.algorithms.deep_cfr."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import parameterized
import tensorflow as tf

from open_spiel.python import policy
from open_spiel.python.algorithms import deep_cfr_tf2
from open_spiel.python.algorithms import exploitability
import pyspiel


class DeepCFRTest(parameterized.TestCase):

  @parameterized.parameters('leduc_poker', 'kuhn_poker', 'liars_dice')
  def test_deep_cfr_runs(self, game_name):
    game = pyspiel.load_game(game_name)
    deep_cfr_solver = deep_cfr_tf2.DeepCFRSolver(
        game,
        policy_network_layers=(8, 4),
        advantage_network_layers=(4, 2),
        num_iterations=2,
        num_traversals=2,
        learning_rate=1e-3,
        batch_size_advantage=8,
        batch_size_strategy=8,
        memory_capacity=1e7)
    deep_cfr_solver.solve()

  def test_matching_pennies_3p(self):
    # We don't expect Deep CFR to necessarily converge on 3-player games but
    # it's nonetheless interesting to see this result.
    game = pyspiel.load_game_as_turn_based('matching_pennies_3p')
    deep_cfr_solver = deep_cfr_tf2.DeepCFRSolver(
        game,
        policy_network_layers=(16, 8),
        advantage_network_layers=(32, 16),
        num_iterations=2,
        num_traversals=2,
        learning_rate=1e-3,
        batch_size_advantage=8,
        batch_size_strategy=8,
        memory_capacity=1e7)
    deep_cfr_solver.solve()
    conv = exploitability.nash_conv(
        game,
        policy.tabular_policy_from_callable(
            game, deep_cfr_solver.action_probabilities))
    print('Deep CFR in Matching Pennies 3p. NashConv: {}'.format(conv))


if __name__ == '__main__':
  tf.test.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Tests for Python Predator-Prey game."""

import math
from absl.testing import absltest
from absl.testing import parameterized
import numpy as np
import numpy.testing as npt
from open_spiel.python.mfg.games import predator_prey
import pyspiel


class MFGPredatorPreyGameTest(parameterized.TestCase):

  def test_load(self):
    game = pyspiel.load_game('python_mfg_predator_prey')
    game.new_initial_state_for_population(0)
    game.new_initial_state_for_population(1)

  @parameterized.parameters(
      {
          'geometry': predator_prey.Geometry.SQUARE,
          'expected_pos': np.array([0, 4])
      },
      {
          'geometry': predator_prey.Geometry.TORUS,
          'expected_pos': np.array([0, 0])
      },
  )
  def test_dynamics(self, geometry, expected_pos):
    game = pyspiel.load_game('python_mfg_predator_prey', {'geometry': geometry})
    state = game.new_initial_state_for_population(2)
    # Initial chance node.
    self.assertEqual(state.current_player(), pyspiel.PlayerId.CHANCE)
    self.assertLen(state.chance_outcomes(), 1)
    self.assertEqual(state.chance_outcomes()[0][0],
                     predator_prey.pos_to_merged(np.array([0, 4]), state.size))
    state.apply_action(state.chance_outcomes()[0][0])
    self.assertEqual(state.current_player(), 2)
    npt.assert_array_equal(state.pos, [0, 4])
    self.assertEqual(state._action_to_string(player=2, action=2), '[0 1]')
    state.apply_action(2)
    npt.assert_array_equal(state.pos, expected_pos)

  def test_create_with_params(self):
    game = pyspiel.load_game('python_mfg_predator_prey(horizon=100,size=20)')
    self.assertEqual(game.size, 20)
    self.assertEqual(game.horizon, 100)

  @parameterized.parameters(
      {'population': 0},
      {'population': 1},
      {'population': 2},
  )
  def test_random_game(self, population):
    """Tests basic API functions."""
    horizon = 10
    size = 20
    game = predator_prey.MFGPredatorPreyGame(params={
        'horizon': horizon,
        'size': size,
    })
    pyspiel.random_sim_test(
        game,
        num_sims=10,
        serialize=False,
        verbose=True,
        mean_field_population=population)

  @parameterized.parameters(
      {
          'reward_matrix':
              np.array([
                  [0, 1],  #
                  [-1, 0]
              ]),
          'population':
              0,
          'players': 2,
          'initial_pos':
              np.array([0, 0]),
          'distributions': [
              # First pop.
              np.array([
                  [1, 0],  #
                  [0, 0]
              ]),
              # Second pop.
              np.array([
                  [0.5, 0.1],  #
                  [0, 0.9]
              ])
          ],
          'expected_rewards':
              np.array([
                  -math.log(1 + 1e-25) + 0.5,  #
                  -math.log(0.5 + 1e-25) - 1,
              ]),
      },
      {
          'reward_matrix':
              np.array([
                  [0, -1, 0.5],  #
                  [0.5, 0, -1],  #
                  [-0.5, 1, 0],
              ]),
          'population':
              2,
          'players': 3,
          'initial_pos':
              np.array([1, 1]),
          'distributions': [
              # First pop.
              np.array([
                  [0.1, 0.2],  #
                  [0.3, 0.4]
              ]),
              # Second pop.
              np.array([
                  [0.2, 0.1],  #
                  [0.1, 0.6]
              ]),
              # Third pop.
              np.array([
                  [0, 0.1],  #
                  [0.1, 0.8]
              ]),
          ],
          'expected_rewards':
              np.array([
                  -math.log(0.4 + 1e-25) - 0.6 + 0.5 * 0.8,
                  -math.log(0.6 + 1e-25) + 0.5 * 0.4 - 0.8,
                  -math.log(0.8 + 1e-25) - 0.5 * 0.4 + 0.6,
              ]),
      },
  )
  def test_rewards(self, reward_matrix, players, population, initial_pos,
                   distributions, expected_rewards):
    game = pyspiel.load_game(
        'python_mfg_predator_prey', {
            'size': 2,
            'reward_matrix': ' '.join(str(v) for v in reward_matrix.flatten()),
            'players': players
        })
    state = game.new_initial_state_for_population(population)
    # Initial chance node.
    self.assertEqual(state.current_player(), pyspiel.PlayerId.CHANCE)
    state.apply_action(predator_prey.pos_to_merged(initial_pos, state.size))
    self.assertEqual(state.current_player(), population)
    npt.assert_array_equal(state.pos, initial_pos)
    state.apply_action(state._NEUTRAL_ACTION)
    npt.assert_array_equal(state.pos, initial_pos)
    self.assertEqual(state.current_player(), pyspiel.PlayerId.CHANCE)
    state.apply_action(state._NEUTRAL_ACTION)
    self.assertEqual(state.current_player(), pyspiel.PlayerId.MEAN_FIELD)

    # Maps states (in string representation) to their proba.
    dist = {}
    for x in range(state.size):
      for y in range(state.size):
        for pop in range(len(reward_matrix)):
          state_str = state.state_to_str(
              np.array([x, y]),
              state.t,
              pop,
              player_id=pyspiel.PlayerId.MEAN_FIELD)
          dist[state_str] = distributions[pop][y][x]
    support = state.distribution_support()
    state.update_distribution([dist[s] for s in support])

    # Decision node where we get a reward.
    self.assertEqual(state.current_player(), population)
    npt.assert_array_equal(state.rewards(), expected_rewards)


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for open_spiel.python.utils.lru_cache."""

from absl.testing import absltest

from open_spiel.python.utils import lru_cache


class LruCacheTest(absltest.TestCase):

  def test_lru_cache(self):
    cache = lru_cache.LRUCache(4)

    self.assertEmpty(cache)

    info = cache.info()
    self.assertEqual(info.hits, 0)
    self.assertEqual(info.misses, 0)
    self.assertEqual(info.size, 0)
    self.assertEqual(info.max_size, 4)
    self.assertEqual(info.usage, 0)
    self.assertEqual(info.hit_rate, 0)

    self.assertIsNone(cache.get(1))

    cache.set(13, "13")
    self.assertLen(cache, 1)

    self.assertIsNone(cache.get(1))

    self.assertEqual(cache.get(13), "13")

    cache.set(14, "14")
    cache.set(15, "15")
    cache.set(16, "16")

    self.assertLen(cache, 4)

    cache.set(17, "17")

    self.assertLen(cache, 4)

    self.assertIsNone(cache.get(13))  # evicted
    self.assertTrue(cache.get(14))

    self.assertLen(cache, 4)

    cache.set(18, "18")

    self.assertIsNone(cache.get(15))  # evicted
    self.assertTrue(cache.get(14))  # older but more recently used

    info = cache.info()
    self.assertEqual(info.usage, 1)

    cache.clear()

    self.assertIsNone(cache.get(18))  # evicted

    self.assertEqual(cache.make(19, lambda: "19"), "19")
    self.assertEqual(cache.get(19), "19")
    self.assertEqual(cache.make(19, lambda: "20"), "19")

if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Python spiel example to get all the states in the game."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl import app
from absl import flags

# pylint: disable=unused-import
from open_spiel.python import games
from open_spiel.python.algorithms import get_all_states
from open_spiel.python.mfg import games as mfg_games
import pyspiel

FLAGS = flags.FLAGS

flags.DEFINE_string("game", "tic_tac_toe", "Name of the game")
flags.DEFINE_integer("players", None, "Number of players")
flags.DEFINE_integer("depth_limit", -1, "Depth limit to stop at")
flags.DEFINE_bool("include_terminals", True, "Include terminal states?")
flags.DEFINE_bool("include_chance_states", True, "Include chance states?")


def main(_):
  games_list = pyspiel.registered_games()
  print("Registered games:")
  for game in games_list:
    print(" ", game.short_name)
  print()

  print("Creating game:", FLAGS.game)
  params = {}
  if FLAGS.players is not None:
    params["players"] = FLAGS.players
  game = pyspiel.load_game(FLAGS.game, params)

  print("Getting all states; depth_limit = {}".format(FLAGS.depth_limit))
  all_states = get_all_states.get_all_states(game, FLAGS.depth_limit,
                                             FLAGS.include_terminals,
                                             FLAGS.include_chance_states)

  count = 0
  for state in all_states:
    print(state)
    count += 1

  print()
  print("Total: {} states.".format(count))


if __name__ == "__main__":
  app.run(main)



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for metrics."""

import glob
import os

from absl.testing import absltest
from absl.testing import parameterized

from open_spiel.python.utils import metrics


class MetricsTest(parameterized.TestCase):

  @parameterized.parameters((True,), (False,))
  def test_create(self, just_logging: bool):
    logdir = self.create_tempdir()
    # Create the writer.
    writer = metrics.create_default_writer(
        logdir.full_path, just_logging=just_logging)
    self.assertIsInstance(writer, metrics.metric_writers.MultiWriter)

    # Write some metrics.
    writer.write_hparams({"param1": 1.0, "param2": 2.0})
    for step in range(5):
      writer.write_scalars(step, {"value": step * step})

    metrics.write_values(writer, 5, {
        "scalar": 1.23,
        "text": metrics.Text(value="foo")
    })
    # Flush the writer.
    writer.flush()

    # Check that the summary file exists if not just logging.
    self.assertLen(
        glob.glob(os.path.join(logdir.full_path, "events.out.tfevents.*")),
        0 if just_logging else 1)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Test that Python numpy arrays can be passed to C++ Eigen library."""

import time

from absl.testing import absltest
import numpy as np

import pyspiel_eigen_test


class PyEigenTest(absltest.TestCase):

  def test_square_matrix_elements(self):
    x = np.array([[1, 2], [3, 4]]).astype(float)
    expected = np.array([[1, 2], [3, 4]]) ** 2
    actual = pyspiel_eigen_test.square(x)
    np.testing.assert_array_equal(expected, actual)

  def test_transpose_and_square_matrix_elements(self):
    x = np.array([[1, 2], [3, 4]]).astype(float)
    x = x.transpose()
    expected = np.array(
        [[1, 9],
         [4, 16]])
    actual = pyspiel_eigen_test.square(x)
    np.testing.assert_array_equal(expected, actual)

  def test_transpose_then_slice_and_square_matrix_elements(self):
    x = np.array([[1, 2], [3, 4]]).astype(float)
    x = x.transpose()
    expected = np.array([[9], [16]])
    actual = pyspiel_eigen_test.square(x[0:, 1:])
    np.testing.assert_array_equal(expected, actual)

  def test_square_vector_elements(self):
    x = np.array([1, 2, 3]).astype(float)
    expected = np.array([[1], [4], [9]])
    actual = pyspiel_eigen_test.square(x)
    np.testing.assert_array_equal(expected, actual)

  def test_allocate_cxx(self):
    actual = pyspiel_eigen_test.matrix()
    expected = np.array([[1, 2], [3, 4]])
    np.testing.assert_array_equal(expected, actual)

  def test_flags_copy_or_reference(self):
    # A test implementing
    # https://pybind11.readthedocs.io/en/stable/advanced/cast/eigen.html#returning-values-to-python
    start = time.time()
    a = pyspiel_eigen_test.BigMatrix()
    print("Alloc: ", time.time() - start)

    start = time.time()
    m = a.get_matrix()
    print("Ref get: ", time.time() - start)
    self.assertTrue(m.flags.writeable)
    self.assertFalse(m.flags.owndata)

    start = time.time()
    v = a.view_matrix()
    print("Ref view: ", time.time() - start)
    self.assertFalse(v.flags.writeable)
    self.assertFalse(v.flags.owndata)

    start = time.time()
    c = a.copy_matrix()
    print("Copy: ", time.time() - start)
    self.assertTrue(c.flags.writeable)
    self.assertTrue(c.flags.owndata)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Replay buffer of fixed size with a FIFI replacement policy."""

import random


class ReplayBuffer(object):
  """ReplayBuffer of fixed size with a FIFO replacement policy.

  Stored transitions can be sampled uniformly.

  The underlying datastructure is a ring buffer, allowing 0(1) adding and
  sampling.
  """

  def __init__(self, replay_buffer_capacity):
    self._replay_buffer_capacity = replay_buffer_capacity
    self._data = []
    self._next_entry_index = 0

  def add(self, element):
    """Adds `element` to the buffer.

    If the buffer is full, the oldest element will be replaced.

    Args:
      element: data to be added to the buffer.
    """
    if len(self._data) < self._replay_buffer_capacity:
      self._data.append(element)
    else:
      self._data[self._next_entry_index] = element
      self._next_entry_index += 1
      self._next_entry_index %= self._replay_buffer_capacity

  def sample(self, num_samples):
    """Returns `num_samples` uniformly sampled from the buffer.

    Args:
      num_samples: `int`, number of samples to draw.

    Returns:
      An iterable over `num_samples` random elements of the buffer.

    Raises:
      ValueError: If there are less than `num_samples` elements in the buffer
    """
    if len(self._data) < num_samples:
      raise ValueError("{} elements could not be sampled from size {}".format(
          num_samples, len(self._data)))
    return random.sample(self._data, num_samples)

  def reset(self):
    """Resets the contents of the replay buffer."""
    self._data = []
    self._next_entry_index = 0

  def __len__(self):
    return len(self._data)

  def __iter__(self):
    return iter(self._data)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the heuristic_payoff_table library."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl import logging
from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from open_spiel.python.egt import heuristic_payoff_table
from open_spiel.python.egt import utils
import pyspiel


class ModuleLevelTest(absltest.TestCase):

  def test__multinomial_coefficients(self):
    distributions = np.asarray([
        [2, 0],
        [1, 1],
        [1, 0],
    ])
    coefficients = heuristic_payoff_table._multinomial_coefficients(
        distributions)

    np.testing.assert_array_equal([1., 2., 1.], coefficients)

    distributions = np.asarray([
        [3, 0],
        [2, 1],
        [1, 2],
        [0, 3],
    ])
    coefficients = heuristic_payoff_table._multinomial_coefficients(
        distributions)
    np.testing.assert_array_equal([1., 3., 3., 1.], coefficients)

    distributions = np.asarray([
        [2, 0, 0],
        [0, 2, 0],
        [0, 0, 2],
        [1, 1, 0],
        [1, 0, 1],
        [0, 1, 1],
    ])
    coefficients = heuristic_payoff_table._multinomial_coefficients(
        distributions)
    np.testing.assert_array_equal([1., 1., 1., 2., 2., 2.], coefficients)


class PayoffTableTest(parameterized.TestCase):

  @parameterized.parameters(
      (5, 2),
      (2, 2),
  )
  def test_construction(self, num_players, num_strategies):
    logging.info("Testing payoff table construction.")
    table = heuristic_payoff_table.PayoffTable(num_players, num_strategies)
    num_rows = utils.n_choose_k(num_players + num_strategies - 1, num_players)
    distributions = np.array(
        list(utils.distribute(num_players, num_strategies)))
    payoffs = np.full([int(num_rows), num_strategies], np.nan)
    np.testing.assert_array_equal(
        np.concatenate([distributions, payoffs], axis=1), table())

  def test_from_heuristic_payoff_table(self):
    team_compositions = np.asarray([
        [2, 0],
        [1, 1],
        [0, 2],
    ])
    payoffs = np.asarray([
        [1, 2],
        [3, 4],
        [5, 6],
    ])
    hpt = np.hstack([team_compositions, payoffs])

    table = heuristic_payoff_table.from_heuristic_payoff_table(hpt)
    np.testing.assert_array_equal(team_compositions, table._distributions)
    np.testing.assert_array_equal(payoffs, table._payoffs)
    self.assertEqual(3, table.num_rows)

    distributions = np.asarray([
        [2, 0, 0],
        [0, 2, 0],
        [0, 0, 2],
        [1, 1, 0],
        [1, 0, 1],
        [0, 1, 1],
    ])
    shape = distributions.shape
    payoffs = np.reshape(np.arange(np.prod(shape)), shape)

    hpt = np.hstack([distributions, payoffs])
    table = heuristic_payoff_table.from_heuristic_payoff_table(hpt)
    np.testing.assert_array_equal(distributions, table._distributions)
    np.testing.assert_array_equal(payoffs, table._payoffs)
    self.assertEqual(distributions.shape[0], table.num_rows)

  @parameterized.parameters(("matrix_rps",))
  def test_from_matrix_game(self, game):
    game = pyspiel.load_matrix_game(game)
    payoff_tables = utils.game_payoffs_array(game)
    logging.info("Testing payoff table construction for matrix game.")
    table = heuristic_payoff_table.from_matrix_game(payoff_tables[0])
    print(table())

  @parameterized.parameters((np.array([0.7, 0.2, 0.1]),))
  def test_expected_payoff(self, strategy):
    logging.info("Testing expected payoff for matrix game.")
    game = pyspiel.load_matrix_game("matrix_rps")
    payoff_tables = utils.game_payoffs_array(game)
    table = heuristic_payoff_table.from_matrix_game(payoff_tables[0])
    expected_payoff = table.expected_payoff(strategy)
    print(expected_payoff)
    assert len(expected_payoff) == table._num_strategies

  def test_from_elo_scores(self):
    elo_scores = [800, 400, 400]
    elo_1 = 10**(800 / 400)
    elo_2 = 10**(400 / 400)  # This is also the associated value for player 3.
    expected = np.asarray([
        [2, 0, 0, 1 / 2, 0, 0],
        [0, 2, 0, 0, 1 / 2, 0],
        [0, 0, 2, 0, 0, 1 / 2],
        [1, 1, 0, elo_1 / (elo_1 + elo_2), elo_2 / (elo_1 + elo_2), 0],
        [1, 0, 1, elo_1 / (elo_1 + elo_2), 0, elo_2 / (elo_1 + elo_2)],
        [0, 1, 1, 0, 1 / 2, 1 / 2],
    ])

    htp = heuristic_payoff_table.from_elo_scores(elo_scores)

    np.testing.assert_array_almost_equal(
        utils.sort_rows_lexicographically(expected),
        utils.sort_rows_lexicographically(htp()),
        verbose=True)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Modular implementations of the PSRO meta algorithm.

Allows the use of Restricted Nash Response, Nash Response, Uniform Response,
and other modular matchmaking selection components users can add.

This version works for N player, general sum games.

One iteration of the algorithm consists of:

1) Computing the selection probability vector (or meta-strategy) for current
strategies of each player, given their payoff.
2) [optional] Generating a mask over joint policies that restricts which policy
to train against, ie. rectify the set of policies trained against. (This
operation is designated by "rectify" in the code)
3) From every strategy used, generating a new best response strategy against the
meta-strategy-weighted, potentially rectified, mixture of strategies using an
oracle.
4) Updating meta game matrix with new game results.

"""

import itertools

import numpy as np

from open_spiel.python import policy
from open_spiel.python.algorithms.psro_v2 import abstract_meta_trainer
from open_spiel.python.algorithms.psro_v2 import strategy_selectors
from open_spiel.python.algorithms.psro_v2 import utils


TRAIN_TARGET_SELECTORS = {
    "": None,
    "rectified": strategy_selectors.rectified_selector,
}


class PSROSolver(abstract_meta_trainer.AbstractMetaTrainer):
  """A general implementation PSRO.

  PSRO is the algorithm described in (Lanctot et Al., 2017,
  https://arxiv.org/pdf/1711.00832.pdf ).

  Subsequent work regarding PSRO's matchmaking and training has been performed
  by David Balduzzi, who introduced Restricted Nash Response (RNR), Nash
  Response (NR) and Uniform Response (UR).
  RNR is Algorithm 4 in (Balduzzi, 2019, "Open-ended Learning in Symmetric
  Zero-sum Games"). NR, Nash response, is algorithm 3.
  Balduzzi et Al., 2019, https://arxiv.org/pdf/1901.08106.pdf

  This implementation allows one to modularly choose different meta strategy
  computation methods, or other user-written ones.
  """

  def __init__(self,
               game,
               oracle,
               sims_per_entry,
               initial_policies=None,
               rectifier="",
               training_strategy_selector=None,
               meta_strategy_method="alpharank",
               sample_from_marginals=False,
               number_policies_selected=1,
               n_noisy_copies=0,
               alpha_noise=0.0,
               beta_noise=0.0,
               **kwargs):
    """Initialize the PSRO solver.

    Arguments:
      game: The open_spiel game object.
      oracle: Callable that takes as input: - game - policy - policies played -
        array representing the probability of playing policy i - other kwargs
        and returns a new best response.
      sims_per_entry: Number of simulations to run to estimate each element of
        the game outcome matrix.
      initial_policies: A list of initial policies for each player, from which
        the optimization process will start.
      rectifier: A string indicating the rectifying method. Can be :
              - "" or None: Train against potentially all strategies.
              - "rectified": Train only against strategies beaten by current
                strategy.
      training_strategy_selector: Callable taking (PSROSolver,
        'number_policies_selected') and returning a list of list of selected
        strategies to train from - this usually means copying weights and
        rectifying with respect to the selected strategy's performance (One list
        entry per player), or string selecting pre-implemented methods.
        String value can be:
              - "top_k_probabilites": selects the first
              'number_policies_selected' policies with highest selection
              probabilities.
              - "probabilistic": randomly selects 'number_policies_selected'
                with probabilities determined by the meta strategies.
              - "exhaustive": selects every policy of every player.
              - "rectified": only selects strategies that have nonzero chance of
                being selected.
              - "uniform": randomly selects 'number_policies_selected'
                policies with uniform probabilities.
      meta_strategy_method: String or callable taking a GenPSROSolver object and
        returning two lists ; one list of meta strategies (One list entry per
        player), and one list of joint strategies.
        String value can be:
              - alpharank: AlphaRank distribution on policies.
              - "uniform": Uniform distribution on policies.
              - "nash": Taking nash distribution. Only works for 2 player, 0-sum
                games.
              - "prd": Projected Replicator Dynamics, as described in Lanctot et
                Al.
      sample_from_marginals: A boolean, specifying whether to sample from
        marginal (True) or joint (False) meta-strategy distributions.
      number_policies_selected: Number of policies to return for each player.

      n_noisy_copies: Number of noisy copies of each agent after training. 0 to
        ignore this.
      alpha_noise: lower bound on alpha noise value (Mixture amplitude.)
      beta_noise: lower bound on beta noise value (Softmax temperature.)
      **kwargs: kwargs for meta strategy computation and training strategy
        selection.
    """
    self._sims_per_entry = sims_per_entry
    print("Using {} sims per entry.".format(sims_per_entry))

    self._rectifier = TRAIN_TARGET_SELECTORS.get(
        rectifier, None)
    self._rectify_training = self._rectifier
    print("Rectifier : {}".format(rectifier))

    self._meta_strategy_probabilities = np.array([])
    self._non_marginalized_probabilities = np.array([])

    print("Perturbating oracle outputs : {}".format(n_noisy_copies > 0))
    self._n_noisy_copies = n_noisy_copies
    self._alpha_noise = alpha_noise
    self._beta_noise = beta_noise

    self._policies = []  # A list of size `num_players` of lists containing the
    # strategies of each player.
    self._new_policies = []

    # Alpharank is a special case here, as it's not supported by the abstract
    # meta trainer api, so has to be passed as a function instead of a string.
    if not meta_strategy_method or meta_strategy_method == "alpharank":
      meta_strategy_method = utils.alpharank_strategy

    print("Sampling from marginals : {}".format(sample_from_marginals))
    self.sample_from_marginals = sample_from_marginals

    super(PSROSolver, self).__init__(
        game,
        oracle,
        initial_policies,
        meta_strategy_method,
        training_strategy_selector,
        number_policies_selected=number_policies_selected,
        **kwargs)

  def _initialize_policy(self, initial_policies):
    self._policies = [[] for k in range(self._num_players)]
    self._new_policies = [([initial_policies[k]] if initial_policies else
                           [policy.UniformRandomPolicy(self._game)])
                          for k in range(self._num_players)]

  def _initialize_game_state(self):
    effective_payoff_size = self._game_num_players
    self._meta_games = [
        np.array(utils.empty_list_generator(effective_payoff_size))
        for _ in range(effective_payoff_size)
    ]
    self.update_empirical_gamestate(seed=None)

  def get_joint_policy_ids(self):
    """Returns a list of integers enumerating all joint meta strategies."""
    return utils.get_strategy_profile_ids(self._meta_games)

  def get_joint_policies_from_id_list(self, selected_policy_ids):
    """Returns a list of joint policies from a list of integer IDs.

    Args:
      selected_policy_ids: A list of integer IDs corresponding to the
        meta-strategies, with duplicate entries allowed.

    Returns:
      selected_joint_policies: A list, with each element being a joint policy
        instance (i.e., a list of policies, one per player).
    """
    policies = self.get_policies()

    selected_joint_policies = utils.get_joint_policies_from_id_list(
        self._meta_games, policies, selected_policy_ids)
    return selected_joint_policies

  def update_meta_strategies(self):
    """Recomputes the current meta strategy of each player.

    Given new payoff tables, we call self._meta_strategy_method to update the
    meta-probabilities.
    """
    if self.symmetric_game:
      self._policies = self._policies * self._game_num_players

    self._meta_strategy_probabilities, self._non_marginalized_probabilities = (
        self._meta_strategy_method(solver=self, return_joint=True))

    if self.symmetric_game:
      self._policies = [self._policies[0]]
      self._meta_strategy_probabilities = [self._meta_strategy_probabilities[0]]

  def get_policies_and_strategies(self):
    """Returns current policy sampler, policies and meta-strategies of the game.

    If strategies are rectified, we automatically switch to returning joint
    strategies.

    Returns:
      sample_strategy: A strategy sampling function
      total_policies: A list of list of policies, one list per player.
      probabilities_of_playing_policies: the meta strategies, either joint or
        marginalized.
    """
    sample_strategy = utils.sample_strategy_marginal
    probabilities_of_playing_policies = self.get_meta_strategies()
    if self._rectify_training or not self.sample_from_marginals:
      sample_strategy = utils.sample_strategy_joint
      probabilities_of_playing_policies = self._non_marginalized_probabilities

    total_policies = self.get_policies()
    return sample_strategy, total_policies, probabilities_of_playing_policies

  def _restrict_target_training(self,
                                current_player,
                                ind,
                                total_policies,
                                probabilities_of_playing_policies,
                                restrict_target_training_bool,
                                epsilon=1e-12):
    """Rectifies training.

    Args:
      current_player: the current player.
      ind: Current strategy index of the player.
      total_policies: all policies available to all players.
      probabilities_of_playing_policies: meta strategies.
      restrict_target_training_bool: Boolean specifying whether to restrict
        training. If False, standard meta strategies are returned. Otherwise,
        restricted joint strategies are returned.
      epsilon: threshold below which we consider 0 sum of probabilities.

    Returns:
      Probabilities of playing each joint strategy (If rectifying) / probability
      of each player playing each strategy (Otherwise - marginal probabilities)
    """
    true_shape = tuple([len(a) for a in total_policies])
    if not restrict_target_training_bool:
      return probabilities_of_playing_policies
    else:
      kept_probas = self._rectifier(
          self, current_player, ind)
      # Ensure probabilities_of_playing_policies has same shape as kept_probas.
      probability = probabilities_of_playing_policies.reshape(true_shape)
      probability = probability * kept_probas
      prob_sum = np.sum(probability)

      # If the rectified probabilities are too low / 0, we play against the
      # non-rectified probabilities.
      if prob_sum <= epsilon:
        probability = probabilities_of_playing_policies
      else:
        probability /= prob_sum

      return probability

  def update_agents(self):
    """Updates policies for each player at the same time by calling the oracle.

    The resulting policies are appended to self._new_policies.
    """

    used_policies, used_indexes = self._training_strategy_selector(
        self, self._number_policies_selected)

    (sample_strategy,
     total_policies,
     probabilities_of_playing_policies) = self.get_policies_and_strategies()

    # Contains the training parameters of all trained oracles.
    # This is a list (Size num_players) of list (Size num_new_policies[player]),
    # each dict containing the needed information to train a new best response.
    training_parameters = [[] for _ in range(self._num_players)]

    for current_player in range(self._num_players):
      if self.sample_from_marginals:
        currently_used_policies = used_policies[current_player]
        current_indexes = used_indexes[current_player]
      else:
        currently_used_policies = [
            joint_policy[current_player] for joint_policy in used_policies
        ]
        current_indexes = used_indexes[current_player]

      for i in range(len(currently_used_policies)):
        pol = currently_used_policies[i]
        ind = current_indexes[i]

        new_probabilities = self._restrict_target_training(
            current_player, ind, total_policies,
            probabilities_of_playing_policies,
            self._rectify_training)

        new_parameter = {
            "policy": pol,
            "total_policies": total_policies,
            "current_player": current_player,
            "probabilities_of_playing_policies": new_probabilities
        }
        training_parameters[current_player].append(new_parameter)

    if self.symmetric_game:
      self._policies = self._game_num_players * self._policies
      self._num_players = self._game_num_players
      training_parameters = [training_parameters[0]]

    # List of List of new policies (One list per player)
    self._new_policies = self._oracle(
        self._game,
        training_parameters,
        strategy_sampler=sample_strategy,
        using_joint_strategies=self._rectify_training or
        not self.sample_from_marginals)

    if self.symmetric_game:
      # In a symmetric game, only one population is kept. The below lines
      # therefore make PSRO consider only the first player during training,
      # since both players are identical.
      self._policies = [self._policies[0]]
      self._num_players = 1

  def update_empirical_gamestate(self, seed=None):
    """Given new agents in _new_policies, update meta_games through simulations.

    Args:
      seed: Seed for environment generation.

    Returns:
      Meta game payoff matrix.
    """
    if seed is not None:
      np.random.seed(seed=seed)
    assert self._oracle is not None

    if self.symmetric_game:
      # Switch to considering the game as a symmetric game where players have
      # the same policies & new policies. This allows the empirical gamestate
      # update to function normally.
      self._policies = self._game_num_players * self._policies
      self._new_policies = self._game_num_players * self._new_policies
      self._num_players = self._game_num_players

    # Concatenate both lists.
    updated_policies = [
        self._policies[k] + self._new_policies[k]
        for k in range(self._num_players)
    ]

    # Each metagame will be (num_strategies)^self._num_players.
    # There are self._num_player metagames, one per player.
    total_number_policies = [
        len(updated_policies[k]) for k in range(self._num_players)
    ]
    number_older_policies = [
        len(self._policies[k]) for k in range(self._num_players)
    ]
    number_new_policies = [
        len(self._new_policies[k]) for k in range(self._num_players)
    ]

    # Initializing the matrix with nans to recognize unestimated states.
    meta_games = [
        np.full(tuple(total_number_policies), np.nan)
        for k in range(self._num_players)
    ]

    # Filling the matrix with already-known values.
    older_policies_slice = tuple(
        [slice(len(self._policies[k])) for k in range(self._num_players)])
    for k in range(self._num_players):
      meta_games[k][older_policies_slice] = self._meta_games[k]

    # Filling the matrix for newly added policies.
    for current_player in range(self._num_players):
      # Only iterate over new policies for current player ; compute on every
      # policy for the other players.
      range_iterators = [
          range(total_number_policies[k]) for k in range(current_player)
      ] + [range(number_new_policies[current_player])] + [
          range(total_number_policies[k])
          for k in range(current_player + 1, self._num_players)
      ]
      for current_index in itertools.product(*range_iterators):
        used_index = list(current_index)
        used_index[current_player] += number_older_policies[current_player]
        if np.isnan(meta_games[current_player][tuple(used_index)]):
          estimated_policies = [
              updated_policies[k][current_index[k]]
              for k in range(current_player)
          ] + [
              self._new_policies[current_player][current_index[current_player]]
          ] + [
              updated_policies[k][current_index[k]]
              for k in range(current_player + 1, self._num_players)
          ]

          if self.symmetric_game:
            # TODO(author4): This update uses ~2**(n_players-1) * sims_per_entry
            # samples to estimate each payoff table entry. This should be
            # brought to sims_per_entry to coincide with expected behavior.
            utility_estimates = self.sample_episodes(estimated_policies,
                                                     self._sims_per_entry)

            player_permutations = list(itertools.permutations(list(range(
                self._num_players))))
            for permutation in player_permutations:
              used_tuple = tuple([used_index[i] for i in permutation])
              for player in range(self._num_players):
                if np.isnan(meta_games[player][used_tuple]):
                  meta_games[player][used_tuple] = 0.0
                meta_games[player][used_tuple] += utility_estimates[
                    permutation[player]] / len(player_permutations)
          else:
            utility_estimates = self.sample_episodes(estimated_policies,
                                                     self._sims_per_entry)
            for k in range(self._num_players):
              meta_games[k][tuple(used_index)] = utility_estimates[k]

    if self.symmetric_game:
      # Make PSRO consider that we only have one population again, as we
      # consider that we are in a symmetric game (No difference between players)
      self._policies = [self._policies[0]]
      self._new_policies = [self._new_policies[0]]
      updated_policies = [updated_policies[0]]
      self._num_players = 1

    self._meta_games = meta_games
    self._policies = updated_policies
    return meta_games

  def get_meta_game(self):
    """Returns the meta game matrix."""
    return self._meta_games

  @property
  def meta_games(self):
    return self._meta_games

  def get_policies(self):
    """Returns a list, each element being a list of each player's policies."""
    policies = self._policies
    if self.symmetric_game:
      # For compatibility reasons, return list of expected length.
      policies = self._game_num_players * self._policies
    return policies

  def get_and_update_non_marginalized_meta_strategies(self, update=True):
    """Returns the Nash Equilibrium distribution on meta game matrix."""
    if update:
      self.update_meta_strategies()
    return self._non_marginalized_probabilities

  def get_strategy_computation_and_selection_kwargs(self):
    return self._strategy_computation_and_selection_kwargs
